@@ -6,7 +6,7 @@ perlcassa - Perl Client for Apache Cassandra
 
 =head1 VERSION
 
-v0.03
+v0.04
 
 =head1 SYNOPSIS
 
@@ -126,7 +126,7 @@ limitations under the License.
 use strict;
 use warnings;
 
-our $VERSION = '0.02';
+our $VERSION = '0.04';
 
 use perlcassa::Client qw/setup close_conn client_setup/;
 
@@ -153,6 +153,11 @@ our %validation_map = (
 	'UUIDType'	=> 'S'
 );
 
+# Get the time in microsecond precision
+sub ustime {
+    return int(1000*1000*time());
+}
+
 sub new() {
 	my ($class, %opt) = @_;
 
@@ -178,7 +183,8 @@ sub new() {
 		request_count => 0,
 		availablehosts => {},
 		failure_thread_running => 0,
-		max_retry => $opt{max_retry} || 2 # max number of times to try and connect for this request
+		max_retry => $opt{max_retry} || 2, # max number of times to try and connect for this request
+		donotpack => $opt{donotpack} || 0 #will insert with raw values and not autopack values
 	}, $class;
 
 	# generate the inital randomized server list
@@ -305,13 +311,12 @@ sub execute() {
 sub describe_all_columnfamilies() {
 	my ($self, %opts)  = @_;
 
-	my $keyspace = $opts{keyspace};
+	my $keyspace = $opts{keyspace} || $self->{keyspace};
 	
 	$self->client_setup('keyspace' => $keyspace);
 	my $client = $self->{client};
 
 	my $keyspacedesc = $client->describe_keyspace($keyspace);
-
 	my $columnfamilydesc = $keyspacedesc->{cf_defs};
 
 	my %return;
@@ -462,7 +467,8 @@ sub insert() {
 	};
 
 	if ($@) {
-		die("[ERROR] Insert was unsucessful $@");
+		my %error = %{$@};
+		die("[ERROR] Insert was unsucessful: $error{why}");
 	}
 }
 
@@ -487,7 +493,8 @@ sub remove() {
 	};
 
 	if ($@) {
-		die("[ERROR] Remove was unsuccessful $@");
+		my %error = %{$@};
+		die("[ERROR] Remove was unsuccessful: $error{why}");
 	}
 
 }
@@ -527,7 +534,8 @@ sub add() {
 	};
 
 	if ($@) {
-		die("[ERROR] add failed. $@");
+		my %error = %{$@};
+		die("[ERROR] Add Failed: $error{why}");
 	}
 }
 
@@ -575,7 +583,7 @@ sub _call() {
 		unless ($thrift_operation eq 'add') {
 			# pack the value to comply with the default column family validator on this column family
 			my %valuehash = ('values' => [$value]);
-			$packedvalue = $self->_pack_values(\%valuehash, $columnfamily, 'value');
+			$packedvalue = $self->_pack_values(\%valuehash, $columnfamily, 'value', $name);
 		}
 	}
 
@@ -591,15 +599,15 @@ sub _call() {
 		$client->add($packedkey, $column_parent, $countercolumn, $consistencylevel);
 	} elsif ($thrift_operation eq 'remove') { 
 
-		$client->remove($packedkey, $column_parent, time, $consistencylevel);
+		$client->remove($packedkey, $column_parent, ustime, $consistencylevel);
 	} else {
 		my @mutations;
 
 		# first create the column	
 		my $column = new Cassandra::Column();
 		$column->{name} = $name;
-		$column->{value} = $value;
-		$column->{timestamp} = time;
+		$column->{value} = $packedvalue;
+		$column->{timestamp} = ustime;
 
 		if(defined($ttl)){
 			$column->{ttl} = $ttl; 
@@ -639,13 +647,17 @@ sub bulk_insert() {
 	my %columns = %$columns;
 	my %packedbulk;
 	foreach my $value (sort(keys(%columns))) {
-		my %valuehash = ('values' => [$value]);
-		my $packedvalue = $self->_pack_values(\%valuehash, $columnfamily, 'value');
+		if ($self->{donotpack} == 1) {
+			$packedbulk{$value} = @{$columns{$value}}[0];
+		} else {
+			my %valuehash = ('values' => \@{$columns{$value}});
+			my $packedvalue = $self->_pack_values(\%valuehash, $columnfamily, 'value', $value);
 
-		my %columnhash = ('values' => \@{$columns{$value}});
-		my $packedname = $self->_pack_values(\%columnhash, $columnfamily, 'column');
-
-		$packedbulk{$packedname} = $packedvalue;
+			my %columnhash = ('values' => [$value]);
+			my $packedname = $self->_pack_values(\%columnhash, $columnfamily, 'column');
+		
+			$packedbulk{$packedname} = $packedvalue;
+		}
 	}
 
 	my @mutations;
@@ -655,7 +667,7 @@ sub bulk_insert() {
 		my $column = new Cassandra::Column();
 		$column->{name} = $key;
 		$column->{value} = $packedbulk{$key};
-		$column->{timestamp} = time;
+		$column->{timestamp} = ustime;
 
 		# create a ColumnOrSuperColumn object to put the Column in
 		my $c_or_sc = new Cassandra::ColumnOrSuperColumn();
@@ -731,37 +743,184 @@ sub get() {
 	my $packedkey = $self->_pack_values(\%keyhash, $column_family, 'key');
 
 	my $res = $self->_get_column($column_family, $column, $key, $consistencylevel);
+	my $data = $res->{column} || $res->{counter_column};
 
-	my $value = $self->_unpack_value(packedstr => $res->{column}->{value}, mode => 'value_validation');
+	my $value = $self->_unpack_value(
+		name => [$res->{column}->{name}],
+		packedstr => $data->{value},
+		columnfamily => $column_family,
+		mode => 'value_validation'
+	);
 
 	return $value;
 }
 
+#####################################################################################################
+# multiget_slice allows you to pull out a keys/columns from cassandra. The client will deserialize the data
+# and will return a hash ref:
+# { key => { column => value, }, ... }
+#
+# $obj->multiget_slice(
+#	'columnfamily'  => 'myCF', #optional if provided in object creation
+#	'keyspace'	=> 'myKeyspace', #optional if provided in object creation
+#	'keys'		=> ['key', ...],
+#	'columnnames'	=> ['column' , ...]
+# );
+#
+# $obj->multiget_slice(
+#	'columnfamily'  => 'myCF', #optional if provided in object creation
+#	'keyspace'	=> 'myKeyspace', #optional if provided in object creation
+#	'keys'		=> ['key', ...],
+#	'columnstart'	=> 'abc',
+#	'columnfinish'	=> 'zxy',
+#	'reversed'	=> 0, #optional, defaults to 0, specify 1 to reverse results
+#	'count'		=> 100 #optional, defaults to 100, specify number of columns to return in query
+# );
+#####################################################################################################
+sub multiget_slice {
+	my ($self, %opts) = @_;
+
+	my $column_family = $opts{columnfamily} || $self->{columnfamily};
+	my $keyspace = $opts{keyspace} || $self->{keyspace};
+	my $consistencylevel = $opts{consistency_level} || $self->{read_consistency_level};
+	my $columns = $opts{columnnames};
+
+	$self->client_setup('keyspace' => $keyspace, 'columnfamily' => $column_family);
+
+	my @keys = @{ $opts{keys} // [] };
+	my @keys_packed;
+	foreach(@keys) {
+        	push @keys_packed, $self->_pack_values({values => [$_]}, $column_family, 'key');
+	}
+
+	# pack the column names unless object created specifying we shouldn't pack the values
+	if ($self->{donotpack} == 1) {
+		unless (ref($columns) eq "HASH") {
+			# deal with a scalar being passed instead of an array
+			my %columnnamehash = ('values' => [$columns]);
+			$columns = $self->_pack_values(\%columnnamehash, $column_family, 'column');
+		} else {
+			$columns = $self->_pack_values($columns, $column_family, 'column');
+		}
+	}
+
+	my $column_parent = new Cassandra::ColumnPath();
+	$column_parent->{column_family} = $column_family;
+	$column_parent->{column} = $columns;
+
+	my $slice_range = new Cassandra::SliceRange();
+	$slice_range->{start} = $opts{columnstart};
+	$slice_range->{finish} = $opts{columnfinish};
+	if (defined($opts{reversed}) && $opts{reversed} == 1) {
+		$slice_range->{reversed} = 1;
+	}
+	
+	$slice_range->{count} = $opts{count} // 100;
+
+	# you can either have a slice_range OR an array of column_names
+	if (defined($opts{columnstart}) && defined($opts{columnnames})) {
+		die('multiget_slice can either have a column range specified (start / finish) OR an array of column names');
+	}
+
+	my $predicate;
+	if (defined($opts{columnstart}) && defined($opts{columnfinish})) {
+		$predicate = Cassandra::SlicePredicate->new({
+			slice_range     => $slice_range
+		});
+	} elsif (defined(@{$opts{columnnames}}[0])) {
+		$predicate = Cassandra::SlicePredicate->new({
+			column_names	=> $opts{columnnames}
+		});
+	} else {
+		die('multiget_slice must be called with either an array of columnnames OR a start and finish column name'); 
+	}
+
+
+	my $res = $self->_do_multiget_slice_with_timeout(sub {
+		$self->{client}->multiget_slice(
+		\@keys,
+		$column_parent,
+		$predicate,
+		$consistencylevel,
+		);
+	});
+
+	my $r = {};
+	while(my($k,$v) = each %{ $res // {} }) {
+		my %cols;
+		foreach(@{ $v // [] }) {
+			my $data = $_->{column} || $_->{counter_column};
+			$cols{$data->{name}} = $self->_unpack_value(
+				name => [$data->{name}],
+				packedstr => $data->{value},
+				columnfamily => $column_family,
+				mode => 'value_validation'
+			);
+		}
+		$r->{$k}  = \%cols;
+	}
+
+	return $r;
+}
+
+sub _do_multiget_slice_with_timeout {
+	my ($self, $func, $args)    = @_;
+
+	if (defined($self->{timeout})) {
+		local $SIG{ALRM} = sub { die "Get Timed Out"; };
+		my $alarm = alarm($self->{timeout});
+		&$func(@{ $args // [] });
+		alarm($alarm);
+	} else {
+		&$func(@{ $args // [] });
+	}
+}
+
 #####################################################################################################
 # _pack_values() takes a string and packs the values as determined by the ComparatorType on the columnfamily
 # The ComparatorType is pulled from Cassandra by default, future versions will let you manually specify
-# the validation classes
+# the validation classes. $name should be the columnname and only needs to be passed when you are packing a value
 #
 # my %composite = (
 # 	'values'     => ['test1', 0, 'bytes2', time],
+#	'column_comparators' => ['UTF8Type','IntegerType','UTF8Type','LongType'] # optional, manually specify validators on a per pack basis
 # );
 #
-# _pack_values(\%composite, 'mycf'); 
+# _pack_values(\%composite, 'mycf');
 #####################################################################################################
 sub _pack_values() {
-	my ($self, $composite, $columnfamily, $type) = @_;
+	my ($self, $composite, $columnfamily, $type, $columnname) = @_;
 	my %composite = %$composite;
 
-	# if a array of validation classes has been passed in with the name hash, use that, otherwise, determine it from the keyspace definition
+	# if metadata was specified on the columnfamily we need to check if the column being packed
+	# has a different validation class specified than the default validation class
+	my %metadata_validators = %{$self->{metadata_validation}{$columnfamily}};
+
+	# if a array of validation classes has been passed in with the name hash, use that, 
+	# otherwise, determine it from the keyspace definition
 	my @validationComparators;	
 	if (defined($type) && $type eq 'key') {
 		@validationComparators = @{$self->{key_validation}{$columnfamily}};
 	} elsif (defined($type) && $type eq 'value') {
-		@validationComparators = @{$self->{value_validation}{$columnfamily}};
+		# this is a hack to make sure we use the right validation type for now for metadata 
+		# columns with specific values
+		if (defined($columnname) && defined($metadata_validators{$columnname})) {
+			@validationComparators = $metadata_validators{$columnname};
+		} else {
+		# otherwise just use the default value validator for this column family if a specific
+		# value validator for this column name hasn't been passed
+			@validationComparators = @{$self->{value_validation}{$columnfamily}};
+		}
 	} else {
-		@validationComparators = @{$self->{comparators}{$columnfamily}};
+		# use validators provided in composite hash if defined
+		if (defined($composite{column_comparators}) && scalar(@{$composite{column_comparators}}) > 0) {
+			@validationComparators = @{$composite{column_comparatorss}};
+		} else {
+			@validationComparators = @{$self->{comparators}{$columnfamily}};
+		}
 	}
 
+
 	# if this is a query string the logic is a bit different
 	if (defined($composite{start}) && defined($composite{finish})) {
 		my ($startslice, $finishslice);
@@ -867,6 +1026,12 @@ sub _pack_values() {
 
 		my $startquery = pack(join(' ', @startpackoptions),@startpackvalues); 
 		my $finishquery = pack(join(' ', @finishpackoptions),@finishpackvalues);
+
+		if ($self->{debug} == 1) {
+			print STDERR "[DEBUG] HEX String for start query is ".join(" ", map( { sprintf("%02x", ord($_)); } unpack("(a1)*", $startquery))) . "\n";
+			print STDERR "[DEBUG] HEX String for finish query is ".join(" ", map( { sprintf("%02x", ord($_)); } unpack("(a1)*", $finishquery))) . "\n";
+		}
+
 		return (\$startquery, \$finishquery);
 	}
 
@@ -888,6 +1053,7 @@ sub _pack_values() {
 			next;
 		}
 
+
 		$i++; #incremenet our counter for the next item in the array after this pack operation succeeds
 
 		# check what the compositetype validator is supposed to be and pack accordingly
@@ -1049,6 +1215,7 @@ sub _pack_values() {
 			die("unsupported ValidationType specified [$validationtype]");
 		}
 	}
+
 	return pack(join(' ', @packoptions),@packvalues);
 }
 
@@ -1253,6 +1420,7 @@ sub _deserialize_column_array() {
 		# remove expired columns from returned results unless overridden by 'return_expired' in opts:
 		if (defined($result->{column}{timestamp}) && defined($result->{column}{ttl})) {
 			my $expiretime = $result->{column}{timestamp} + $result->{column}{ttl};
+            # XXX TODO may be a bug to compare timestamp with ttl. ttl in seconds, time stamp in milliseconds?
 			if ($expiretime < time && $return_expired != 1) {
 				next;
 			}
@@ -1387,7 +1555,10 @@ sub _call_get_range_slices() {
 	my ($self, %opts) = @_;
 
 	my $client = $self->{client};
-	my @order = @{$opts{order}};
+	my @order;
+	if (defined(@{$opts{order}}[0])) {
+		@order = @{$opts{order}};
+	} 
 
 	my $column_parent = new Cassandra::ColumnParent({column_family => $opts{column_family}});
 	my $slice_range = new Cassandra::SliceRange();
